/*
    Copyright (C) 2014 by Elvis Angelaccio <angelaccioelvis@gmail.com>

	This file is part of Kronometer.

	Kronometer is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	Kronometer is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Kronometer.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"
  
#include <KApplication>
#include <KLocale>
#include <KActionCollection>
#include <KStatusBar>
#include <KConfigDialog>
#include <KFileDialog>
#include <KMessageBox>
#include <KIO/NetAccess>
#include <KSaveFile>

#include <QHeaderView>

#include "settings.h"

namespace
{
    const char START_MSG[] = "&Start";
    const char PAUSE_MSG[] = "&Pause";
    const char RESET_MSG[] = "&Reset";
    const char RESUME_MSG[] = "Re&sume";
    const char LAP_MSG[] = "&Lap";

	const QString START_KEY = "start";
	const QString PAUSE_KEY = "pause";
	const QString RESET_KEY = "reset";
	const QString LAP_KEY = "lap";
	
    const char RUNNING_MSG[] = "Running...";
    const char PAUSED_MSG[] = "Paused";
    const char INACTIVE_MSG[] = "Inactive";

    const QString WINDOW_TITLE = "Kronometer";       /** Default Window title */
    const QString QT_PLACE_HOLDER = "[*]";           /** Qt standard placeholder for setWindowModified() */

    // kronometerui.rc states
    const QString INACTIVE_STATE = "inactive";
    const QString RUNNING_STATE = "running";
    const QString PAUSED_STATE = "paused";
    const QString PAUSED_FILE_STATE = "pausedFile";  /** An open file has been paused */
}

 
MainWindow::MainWindow(QWidget *parent, const QString& file) : KXmlGuiWindow(parent), unsavedTimes(false)
{
    stopwatch = new QStopwatch(this);
    stopwatchDisplay = new QTimeDisplay(this);
    connect(stopwatch, SIGNAL(time(QTime)), stopwatchDisplay, SLOT(time(QTime)));  // bind stopwatch to its display
    setCentralWidget(stopwatchDisplay);

    setupDock();
    setupStatusBar();
	setupActions();
	loadSettings();

    setWindowTitle(WINDOW_TITLE + QT_PLACE_HOLDER);

    if (not file.isEmpty())
        openFile(file);
}

bool MainWindow::queryClose()
{
    if (stopwatch->isInactive() or not KronometerConfig::askOnExit())
        return true;  // exit without ask

    if (stopwatch->isRunning())
    {
        stopwatch->pause();
        paused();
    }

    int buttonCode;

    if (fileName.isEmpty())
    {
        buttonCode = KMessageBox::warningYesNoCancel(this, i18n("Save times on a new file?"));

        switch (buttonCode)
        {
            case KMessageBox::Yes :
              qDebug() << "chiamo saveFileAs() perchè non ho nessun file";
              saveFileAs();
              return true;  // TODO: return false if saving fails
            case KMessageBox::No :
              return true;
            default: // cancel
              return false;
        }
    }

    else if (unsavedTimes)
    {
        QFileInfo fileInfo(fileName);
        QString msg = "Save times on file " + fileInfo.fileName() + "?";
        buttonCode = KMessageBox::warningYesNoCancel(this, i18n(msg.toAscii()));

        switch (buttonCode)
        {
            case KMessageBox::Yes :
              // save document here. If saving fails, return false;
              qDebug() << "chiamo saveFile() e fa tutto lui!";
              saveFile();
              return true;
            case KMessageBox::No :
              return true;
            default: // cancel
              return false;
        }
    }

    return true;  // there is an open file, but times are already saved.
}
 
void MainWindow::setupDock()
{
    lapModel = new LapModel(this);
    proxyModel = new QSortFilterProxyModel(this);
    proxyModel->setSourceModel(lapModel);

    QDockWidget *lapDock = new QDockWidget(this);
    lapDock->setObjectName("lapDock");
// 	lapDock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
    lapDock->setFeatures(QDockWidget::NoDockWidgetFeatures);
    lapDock->setAllowedAreas(Qt::AllDockWidgetAreas);
    lapDock->setTitleBarWidget(new QWidget(this));  // fake widget to disable titlebar

    lapView = new QTableView(this);
    lapView->setModel(proxyModel);
    lapView->setSelectionBehavior(QAbstractItemView::SelectRows);
    lapView->setGridStyle(Qt::DotLine);
    lapView->verticalHeader()->hide();
    lapView->resizeColumnsToContents();
    lapView->horizontalHeader()->setStretchLastSection(true);
    lapView->setSortingEnabled(true);
    lapView->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Ignored);

    lapDock->setWidget(lapView);
    addDockWidget(Qt::RightDockWidgetArea, lapDock);
}

void MainWindow::setupStatusBar()
{
    statusLabel = new QLabel(this);
    formatLabel = new QLabel(this);

    formatLabel->setToolTip(i18n("Current time format"));
    statusLabel->setToolTip(i18n("Current chronometer status"));

    statusBar()->addWidget(statusLabel);
    statusBar()->addPermanentWidget(formatLabel);
}

void MainWindow::setupActions() 
{
	startAction = new KAction(this);
	pauseAction = new KAction(this);
	resetAction = new KAction(this);
	lapAction = new KAction(this);

	startAction->setIcon(KIcon("player-time"));
	startAction->setShortcut(Qt::Key_Space);
	
	pauseAction->setText(i18n(PAUSE_MSG));  // pauseAction/resetAction have fixed text (startAction don't)
	pauseAction->setIcon(KIcon("media-playback-pause"));
	pauseAction->setShortcut(Qt::Key_Space);
	
	resetAction->setText(i18n(RESET_MSG));
	resetAction->setIcon(KIcon("edit-clear-history"));
	resetAction->setShortcut(Qt::Key_F5);
	
	lapAction->setText(i18n(LAP_MSG));
	lapAction->setIcon(KIcon("chronometer"));
	lapAction->setShortcut(Qt::Key_Return);
	
	actionCollection()->addAction(START_KEY, startAction);
	actionCollection()->addAction(PAUSE_KEY, pauseAction);
	actionCollection()->addAction(RESET_KEY, resetAction);
	actionCollection()->addAction(LAP_KEY, lapAction);

	// triggers for QStopwatch "behavioral" slots
	connect(startAction, SIGNAL(triggered(bool)), stopwatch, SLOT(start()));
	connect(pauseAction, SIGNAL(triggered(bool)), stopwatch, SLOT(pause()));
	connect(resetAction, SIGNAL(triggered(bool)), stopwatch, SLOT(reset()));
	connect(lapAction, SIGNAL(triggered(bool)), stopwatch, SLOT(lap()));
	
	// triggers for LapModel slots
	connect(resetAction, SIGNAL(triggered(bool)), lapModel, SLOT(clear()));
	connect(stopwatch, SIGNAL(lap(QTime)), lapModel, SLOT(lap(QTime)));
	
	// triggers for MainWindow "gui" slots
	connect(startAction, SIGNAL(triggered(bool)), this, SLOT(running()));
	connect(pauseAction, SIGNAL(triggered(bool)), this, SLOT(paused()));
	connect(resetAction, SIGNAL(triggered(bool)), this, SLOT(inactive()));
	connect(lapAction, SIGNAL(triggered(bool)), this, SLOT(updateLapDock()));

    //KStandardAction::quit(kapp, SLOT(quit()), actionCollection());
    KStandardAction::quit(this, SLOT(close()), actionCollection());
	KStandardAction::preferences(this, SLOT(showSettings()), actionCollection());
    KStandardAction::openNew(this, SLOT(newFile()), actionCollection());
    KStandardAction::save(this, SLOT(saveFile()), actionCollection());
    KStandardAction::saveAs(this, SLOT(saveFileAs()), actionCollection());
    KStandardAction::open(this, SLOT(openFile()), actionCollection());

	setupGUI(Default, "kronometerui.rc");
	
    inactive();	// inactive state is the default
}


void MainWindow::loadSettings()
{
    QString timeFormat = setupTimeFormat(
        KronometerConfig::showHours(),
        KronometerConfig::showMinutes(),
        KronometerConfig::showSeconds(),
        KronometerConfig::showTenths(),
        KronometerConfig::showHundredths(),
        KronometerConfig::showMilliseconds()
    );

    lapModel->setTimeFormat(timeFormat);
    stopwatchDisplay->setTimeFormat(timeFormat);
    stopwatchDisplay->setDisplayFont(KronometerConfig::displayFont());
}

void MainWindow::running()
{
    statusLabel->setText(i18n(RUNNING_MSG));

    unsavedTimes = true;
    setWindowModified(unsavedTimes);

    stateChanged(RUNNING_STATE);
}

void MainWindow::paused()
{
    startAction->setText(i18n(RESUME_MSG));
	statusLabel->setText(i18n(PAUSED_MSG));

    if (not fileName.isEmpty())
    {
        stateChanged(PAUSED_FILE_STATE);
    }

    else
    {
        stateChanged(PAUSED_STATE);
    }
}

void MainWindow::inactive()
{
    startAction->setText(i18n(START_MSG));
	statusLabel->setText(i18n(INACTIVE_MSG));

    stateChanged(INACTIVE_STATE);
}

void MainWindow::showSettings()
{
	if (KConfigDialog::showDialog("settings"))
		return;
	
	KConfigDialog* dialog = new KConfigDialog(this, "settings", KronometerConfig::self());
	
	dialog->showButtonSeparator(true);

    KPageWidgetItem *generalPage = dialog->addPage(new GeneralSettings(this), i18n("General settings"));
    generalPage->setIcon(KIcon(KApplication::windowIcon()));

    KPageWidgetItem *fontPage = dialog->addPage(new FontSettings(this), i18n("Font settings"));
    fontPage->setIcon(KIcon("preferences-desktop-font"));

    KPageWidgetItem *savePage = dialog->addPage(new SaveSettings(this), i18n("Save settings"));
    savePage->setIcon(KIcon("document-save"));

    connect(dialog, SIGNAL(settingsChanged(const QString&)), this, SLOT(writeSettings(QString)));
	
    dialog->show();
}

void MainWindow::updateLapDock()
{
	lapView->resizeColumnsToContents();
    lapView->horizontalHeader()->setStretchLastSection(true);
}

void MainWindow::newFile()
{
    MainWindow *window = new MainWindow();
    window->show();
}

void MainWindow::openFile()
{
    QString f = KFileDialog::getOpenFileName();

    MainWindow *window = new MainWindow(nullptr, f);
    window->show();
}

void MainWindow::saveFile()
{
    saveFileAs(fileName);
}

void MainWindow::saveFileAs()
{
    saveFileAs(KFileDialog::getSaveFileName());
}

void MainWindow::writeSettings(const QString& dialogName)
{
	Q_UNUSED(dialogName);
    KronometerConfig::self()->writeConfig();
	
	loadSettings();
}


QString MainWindow::setupTimeFormat(bool hours, bool min, bool sec, bool tenths, bool hundredths, bool msec)
{
    QString timeFormat;
    QString timeFormatMsg;	// status bar message

    if (hours)
    {
        if (min or sec or tenths or hundredths or msec)
        {
            timeFormat = "h:";
            timeFormatMsg = "hours : ";
        }

        else
        {
            timeFormat = "h";
            timeFormatMsg = "hours";
        }
    }

    if (min)
    {
        if (sec or tenths or hundredths or msec)
        {
            timeFormat += "mm:";
            timeFormatMsg += "min : ";
        }

        else
        {
            timeFormat += "mm";
            timeFormatMsg += "min";
        }
    }

    if (sec)
    {
        if (tenths or hundredths or msec)
        {
            timeFormat += "ss.";
            timeFormatMsg += "sec . ";
        }

        else
        {
            timeFormat += "ss";
            timeFormatMsg += "sec";
        }
    }

    if (msec)
    {
        timeFormat += "zzz";
        timeFormatMsg += "msec";
        stopwatch->setGranularity(QStopwatch::MILLISECONDS);
    }

    else if (hundredths)
    {
        timeFormat += "zzz";  // TODO: see if can be += "zz"
        timeFormat = timeFormat.left(timeFormat.size() - 1);
        timeFormatMsg += "hundreths";
        stopwatch->setGranularity(QStopwatch::HUNDREDTHS);
    }

    else if (tenths)
    {
        timeFormat += "zzz";
        timeFormat = timeFormat.left(timeFormat.size() - 2);
        timeFormatMsg += "tenths";
        stopwatch->setGranularity(QStopwatch::TENTHS);
    }

    else
    {
        stopwatch->setGranularity(QStopwatch::SECONDS);
    }

    formatLabel->setText(timeFormatMsg);

    return timeFormat;
}

void MainWindow::saveFileAs(const QString& name)
{
    KSaveFile saveFile(name);
    saveFile.open();

    QDataStream stream(&saveFile);

    stopwatch->serialize(stream);   // save stopwatch time
    stream << *lapModel;            // save laps

    saveFile.finalize();
    saveFile.close();

    fileName = name;

    unsavedTimes = false;
    setWindowModified(unsavedTimes);
}

void MainWindow::openFile(const QString& name)
{
    QString buffer;

    if (KIO::NetAccess::download(name, buffer, this))
    {
        QFile file(buffer);
        file.open(QIODevice::ReadOnly);
        QDataStream stream(&file);

        stopwatch->deserialize(stream);     // load stopwatch time
        stream >> *lapModel;                // load laps
        paused();                           // enter in paused state

        fileName = name;

        KIO::NetAccess::removeTempFile(buffer);
    }

    else
    {
        KMessageBox::error(this, KIO::NetAccess::lastErrorString());
    }

    QFileInfo fileInfo(fileName);
    setWindowTitle(WINDOW_TITLE + " - " + fileInfo.fileName() + QT_PLACE_HOLDER);
}

